# -*- coding: utf-8 -*-
""" Utilities for metric generators."""

import numpy as np
import os

def _read_array(image, dimx, dimy, dimz, dtype):
    v = np.fromfile(image, dtype=dtype, sep="")
    return v.reshape([dimx, dimy, dimz])


def _write_array(A, fileout):
    A.astype('uint8').tofile(fileout)


def _subcube_ids(n_steps, sREV_max_step):
    ids = []
    for l in range(n_steps):
        if (l <= sREV_max_step-1):
            for idx in range(9):
                ids.append((l+1, idx))
        else:
            ids.append((l+1, 0))
    return ids


def make_cut(A, L, cut_size, idx):
    """
    Making subsample cut for a given 3D array. 
    
    **Input:**
     
     	A(np.array): initial 3D array;
     
     	L (tuple (int, int, int)): linear image sizes in x, y and z directions;
     	 
     	cut_size (list [int, int, int]): linear image sizes of subsample;
     	
     	idx (int): index of subcube (0,1,..8). idx = 0 corresponds to the center subsample, idx = 1,..8 corrspond to the corner subsamples.
    """
    if not len(A.shape) == 3:
        raise ValueError("Initial array should have 3 dimensions.")
    if idx < 0 or idx > 8:
        raise ValueError("Index value should be from the set (0,1,..8).")
    if idx == 0:
        return A[int((L[0]-cut_size[0])/2):int((L[0]+cut_size[0])/2), 
                 int((L[1]-cut_size[1])/2):int((L[1]+cut_size[1])/2), 
                 int((L[2]-cut_size[2])/2):int((L[2]+cut_size[2])/2)]
    if idx == 1:
        return A[:cut_size[0], :cut_size[1], :cut_size[2]]
    if idx == 2:
        return A[:cut_size[0], :cut_size[1], L[2]-cut_size[2]:]
    if idx == 3:
        return A[:cut_size[0], L[1]-cut_size[1]:, :cut_size[2]]
    if idx == 4:
        return A[L[0]-cut_size[0]:, :cut_size[1], :cut_size[2]]
    if idx == 5:
        return A[L[0]-cut_size[0]:, L[1]-cut_size[1]:, L[2]-cut_size[2]:]
    if idx == 6:
        return A[:cut_size[0], L[1]-cut_size[1]:, L[2]-cut_size[2]:]
    if idx == 7:
        return A[L[0]-cut_size[0]:, :cut_size[1], L[2]-cut_size[2]:]
    if idx == 8:
        return A[L[0]-cut_size[0]:, L[1]-cut_size[1]:, :cut_size[2]]



# -*- coding: utf-8 -*-
"""Definition of basic vectorizer"""

class BasicVectorizer:
    """
    Base class for vectorizers. (Don't use it directly but derive from it).
    """
    def __init__(self, norm):
        """
        **Input:**
        
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function.
        """
        self.norm = norm



# -*- coding: utf-8 -*-
"""Module for the generation of PNM characteristics using PNM extractor."""
import numpy as np
import time
import os
import porespy as ps
import openpnm as op
import multiprocessing
from functools import partial
from .utils import _subcube_ids, make_cut


def generate_PNM(image, size, n_steps, sREV_max_step, outputdir, n_threads = 1, resolution=1., show_time=False):
    """
    Running PNM extractor for all the selected subsamples.
    
    **Input:**

     	image (numpy.ndarray): 3D array representing the image;
     	
        size (tuple (int, int, int)): linear image sizes in x, y and z directions;
        
     	n_steps (int): number of subsamples selection steps;
     	
     	sREV_max_step (int): maximal step for which sREV and stationarity analysis can be performed;
     	
     	outputdir (str): path to the output folder containing generated data;
     	
        n_threads (int): number of threads used for data generation, default: 1;
        
     	resolution (float): resolution of studied sample, default: 1; 
        
     	show_time (bool): Added to monitor time cost for large images,  default: False. 
    """
    start_time = time.time()
    cut_step = (np.array(size)/n_steps).astype(int)
    cut_sizes = [(cut_step*(i+1)).tolist() for i in range(n_steps-1)]
    cut_sizes.append(size)
    ids = _subcube_ids(n_steps, sREV_max_step)
    cuts = []
    for elem in ids:
        l = elem[0]
        idx = elem[1]
        if l  == n_steps:
            cuts.append(image)
        else:
            cut_size = cut_sizes[l-1]
            cuts.append(make_cut(image, size, cut_size, idx))
    data = zip(ids, cuts)
    pool = multiprocessing.Pool(processes=n_threads)
    results = pool.map(partial(_pnm_for_subsample, outputdir = outputdir, resolution = resolution, show_time = show_time), data)
    pool.close()
    pool.join()
    if show_time:
        print("---total PN data generation time is %s seconds ---" % (time.time() - start_time))

def get_pn_csv(cut, cut_name, outputdir, resolution, show_time):
    """
    Calculation of PNM statistics for a given subsample and writing the result to csv file.
    
    **Input:**
    
        n_threads (int): number of threads used for data generation;

     	cut (numpy.ndarray): 3D array representing a subsample;
     	
     	cut_name (str): name of output file;
     	
     	outputdir (str): path to the output folder containing generated data;
        
     	resolution (float): resolution of studied subsample;
        
        show_time (bool): Added to monitor time cost for large images.
    """
    start_time = time.time()
    cut = cut.astype(bool)
    cut = ~cut
    parallelization = {'cores':1}
    snow_output = ps.networks.snow2(cut, voxel_size = resolution, parallelization = parallelization)
    pn = op.io.network_from_porespy(snow_output.network)
    cut_name1 = os.path.join(outputdir, cut_name)
    op.io.network_to_csv(pn, filename = cut_name1)
    if show_time:
        print(cut_name)
        print("---PNM extractor run time is %s seconds ---" % (time.time() - start_time))

def _pnm_for_subsample(data, outputdir, resolution, show_time):
    l = data[0][0]
    idx = data[0][1]
    cut = data[1]
    cut_name = 'cut'+str(l)+'_'+str(idx)
    get_pn_csv(cut, cut_name, outputdir, resolution, show_time)


# -*- coding: utf-8 -*-
"""Definition of histogram vectorizer"""

import numpy as np
from .basic_vectorizer import BasicVectorizer


class HistVectorizer(BasicVectorizer):
    """
    Class describing vectorizer for histogram-like data.
    """     
    def __init__(self, factor=1, norm=2):
        """
        **Input:**
        
        	factor (int): Defines how many bins are in linear size of one voxel, default: 1;
        	
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function, default: 2.
        """      
        super().__init__(norm)
        self.factor = factor

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        	(list(dtype = float), list(dtype = float), float) - a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        """
        rmax1 = np.ceil(max(v1))
        r1 = [0, rmax1]
        bins = int(self.factor*rmax1)
        hist1, bin_edges1 = np.histogram(v1, bins=bins, range=r1, density=True)
        hist2, bin_edges2 = np.histogram(v2, bins=bins, range=r1, density=True)
        n1 = np.linalg.norm(np.array(hist1), ord=self.norm)
        n2 = np.linalg.norm(np.array(hist2), ord=self.norm)
        delta = 2*np.linalg.norm(np.array(hist1) -
                                 np.array(hist2), ord=self.norm)/(n1 + n2)
        return hist1.tolist(), hist2.tolist(), delta



# -*- coding: utf-8 -*-
"""
Definition of Landscape vectorizer, used in PD vectorization. A persistence landscape is a collection of 1D piecewise-linear functions computed from the rank function associated to the persistence diagram. These piecewise-linear functions are then sampled evenly on a given range and the corresponding vectors of samples are concatenated and returned. See Bubenik, P. (2015). Statistical topological data analysis using persistence landscapes. J. Mach. Learn. Res., 16(1), 77-102.
"""

import numpy as np
from gudhi.representations.vector_methods import Landscape
from .basic_vectorizer import BasicVectorizer


class LandscapeVectorizer(BasicVectorizer):
    """
    Class describing persistence landscape vectorizer.
    """      
    def __init__(self, resolution, num_landcapes=1, norm=2):
        """
        **Input:**
        
        	resolution (int): number of sample for all piecewise-linear functions;
        	 
        	num_landcapes (int): number of piecewise-linear functions to output, default = 1;
        	 
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function, default: 2.
        """         
        super().__init__(norm)
        self.resolution = resolution
        self.num_landscapes = num_landcapes

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        	(list(dtype = float), list(dtype = float), float): a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        """ 
        L1 = Landscape(self.num_landscapes, self.resolution)
        L1.fit([v1])
        range1 = L1.sample_range
        v1 = L1.transform([v1])
        L2 = Landscape(self.num_landscapes,
                       self.resolution, sample_range=range1)
        L2.fit([v2])
        v2 = L2.transform([v2])
        n1 = np.linalg.norm(v1, ord=self.norm)
        n2 = np.linalg.norm(v2, ord=self.norm)
        delta = 2*np.linalg.norm(np.array(v1) -
                                 np.array(v2), ord=self.norm)/(n1 + n2)
        return v1.tolist(), v2.tolist(), delta



# -*- coding: utf-8 -*-
"""Definition of basic metric"""

import numpy as np
import os


class BasicMetric:
    """
    Base class of all metrics. (Don't use it directly but derive from it).
    """    
    def __init__(self, vectorizer, n_threads):
        """
        **Input:**
        
        	vectorizer (subclass of BasicVectorizer): vectorizer to be used for a vector metric;
            
            n_threads (int): number of threads used for data generation.
        """
        self.vectorizer = vectorizer
        self.n_threads = n_threads
        self.directional = False
        self.metric_type = None
        
       
    def read(self, inputdir, step, cut_id):
        """
        Read the metric data generated for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing image;
        	
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index.  
        
        **Output:**
        
        	metric value (float or np.array(dtype='float')).
        """
        if self.directional:
            directions = ('_x', '_y', '_z')
            cut_names = ["cut" + str(step) + "_" + str(cut_id) + direction + ".txt" for direction in directions]
            data = []
            for cut_name in cut_names:
                if inputdir is not None:
                    filein = os.path.join(inputdir, cut_name)
                else:
                    filein = cut_name
                d = np.loadtxt(filein, delimiter=" ", dtype=float)
                if self.metric_type == 'v':
                    data.append(d)
                if self.metric_type == 's':
                    if np.isnan(d) or np.isinf(d):
                        data.append(0)
                    else:
                        data.append(d.item())
            return data
        else:
            cut_name = "cut" + str(step) + "_" + str(cut_id) + ".txt"
            if inputdir is not None:
                filein = os.path.join(inputdir, cut_name)
            else:
                filein = cut_name
            data = np.loadtxt(filein, delimiter=" ", dtype=float)
            if self.metric_type == 's' and np.isnan(data):
                data = np.array(0)
            return data



# -*- coding: utf-8 -*-
"""
Definition of Euler Density I metric. The Euler density here is calculated using the topological properties of considered voxel domain as in Vogel, H. J., Weller, U., & Schlüter, S. (2010). Quantification of soil structure based on Minkowski functions. Computers & Geosciences, 36(10), 1236-1245. See the details in documentation.
"""

from .basic_metric import BasicMetric
from ..generators import _write_array
import os
import time
import imp
import subprocess

class EulerDensityI(BasicMetric):
    """
    Class describing Euler density I metric.
    """     
    def __init__(self, n_threads = 1, show_time=False):
        """
        **Input:**
        
        n_threads (int): number of threads used for data generation, default: 1;
        
        	show_time (bool): flag to monitor time cost for large images, default: False.
        """
        super().__init__(vectorizer=None, n_threads = n_threads)
        self.metric_type = 's'
        self.show_time = show_time

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates Euler density for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """        
        start_time = time.time()
        glob_path = os.getcwd()
        dimx = cut.shape[0]
        dimy = cut.shape[1]
        dimz = cut.shape[2]
        path0 = imp.find_module('revanalyzer')[1]
        jl_path = os.path.join(path0, 'jl', 'euler_density.jl')
        output_path = os.path.join(glob_path, outputdir)
        image_path = os.path.join(output_path, cut_name +'.raw')
        _write_array(cut, image_path)
        file_out = os.path.join(output_path, cut_name +'.txt')
        code = subprocess.call(['julia', jl_path, image_path, str(dimx), str(dimy), str(dimz), file_out])
        if (code != 0):
            raise RuntimeError("Error in julia run occured!")
        os.remove(image_path)
        if self.show_time:
            print("cut ", cut_name, ", run time: ")
            print("--- %s seconds ---" % (time.time() - start_time))



# -*- coding: utf-8 -*-
"""Definition of Porosity metric."""

import numpy as np
import os
from .basic_metric import BasicMetric


class Porosity(BasicMetric):
    """
    Class describing porosity metric.
    """
    def __init__(self, n_threads = 1):
        super().__init__(vectorizer=None, n_threads = n_threads)
        self.metric_type = 's'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates porosity for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        porosity = 1 - np.count_nonzero(cut)/(cut.shape[0]*cut.shape[1]*cut.shape[2])
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        with open(fileout, "w") as f:
            f.write(str(porosity))



# -*- coding: utf-8 -*-
"""
Definition of Silhouette vectorizer, used in PD vectorization. A persistence silhouette is computed by taking a weighted average of the collection of 1D piecewise-linear functions given by the persistence landscapes, and then by evenly sampling this average on a given range. Finally, the corresponding vector of samples is returned. See Chazal, F., et al. (2014). Stochastic convergence of persistence landscapes and silhouettes. In Proceedings of the thirtieth annual symposium on Computational geometry (pp. 474-483).
"""

import numpy as np
from gudhi.representations.vector_methods import Silhouette
from .basic_vectorizer import BasicVectorizer


class SilhouetteVectorizer(BasicVectorizer):
    """
    Class describing persistence silhouette vectorizer.
    """   
    def __init__(self, resolution, n=1, norm=2):
        """
        **Input:**
        
        	resolution (int): number of samples for the weighted average;
        	
        	n (int): power parameter in weighted funtion (d-b)^n, default = 1;
        	 
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function, default: 2.
        """ 
        super().__init__(norm)
        self.resolution = resolution
        self.n = n

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        	(list(dtype = float), list(dtype = float), float): a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        """ 
        S1 = Silhouette(weight=lambda x: (
            x[1] - x[0])**self.n, resolution=self.resolution)
        S1.fit([v1])
        range1 = S1.sample_range
        v1 = S1.transform([v1])
        S2 = Silhouette(weight=lambda x: (
            x[1] - x[0])**self.n, resolution=self.resolution,  sample_range=range1)
        S2.fit([v2])
        v2 = S2.transform([v2])
        n1 = np.linalg.norm(v1, ord=self.norm)
        n2 = np.linalg.norm(v2, ord=self.norm)
        delta = 2*np.linalg.norm(np.array(v1) -
                                 np.array(v2), ord=self.norm)/(n1 + n2)
        return v1.tolist(), v2.tolist(), delta



# -*- coding: utf-8 -*-
"""Definition of CF-based metrics. For the definition of correlation functions (CF) see the documentation."""

from .basic_metric import BasicMetric
from ..generators import _write_array
from ..vectorizers  import CFVectorizer, DirectVectorizer
import numpy as np
import matplotlib.pyplot as plt
import os
import time
import imp
import subprocess


class BasicCFMetric(BasicMetric):
    """
    Base class of CF-based metrics. (Don't use it directly but derive from it).
    """ 
    def __init__(self, vectorizer, n_threads, show_time, normalize):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation;
        	
        	show_time (bool): flag to monitor time cost for large images;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. 
        """        
        super().__init__(vectorizer, n_threads = n_threads)
        self.show_time = show_time
        if normalize == True:
            self.normalize = 1
        else:
            self.normalize = 0

    def generate(self, cut, cut_name, outputdir, method, gendatadir = None):
        """
        Generates CF metric for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder;
        	    
        	method (str): method for generation of cpecific CF. Different in differenent CF-based metrics.
        """          
        start_time = time.time()
        glob_path = os.getcwd()
        dimx = cut.shape[0]
        dimy = cut.shape[1]
        dimz = cut.shape[2]
        path0 = imp.find_module('revanalyzer')[1]
        jl_path = os.path.join(path0, 'jl', 'corfunction_xyz.jl')
        output_path = os.path.join(glob_path, outputdir)
        image_path = os.path.join(output_path, cut_name +'.raw')
        _write_array(cut, image_path)
        file_out = os.path.join(output_path, cut_name)
        code = subprocess.call(['julia', jl_path, image_path, str(dimx), str(dimy), str(dimz), method, str(self.normalize), file_out])
        if (code != 0):
            raise RuntimeError("Error in julia run occured!")
        os.remove(image_path)           
        if self.show_time:
            print(cut_name, ", run time: ")
            print("--- %s seconds ---" % (time.time() - start_time))
        return cut_name + ".txt"

    def show(self, inputdir, step, cut_id, title, cf_type):
        """
        Vizualize CF for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index;
        
            title (str): image title;
            
            cf_type (str): type of CF (directional or probability density).
        """        
        data = self.read(inputdir, step, cut_id)
        _show_cf(data, title, cf_type)

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample;
        
        **Output:**
        
        	Depends on the chosen mode in CFVectorizer.
        
        	If mode = 'all':
        
        		(list(dtype = float), list(dtype = float), float) - a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        
        	If mode = 'max:
        
 			(list(list(dtype = float)), list(list(dtype = float)), list(float)) - a tuple, in which in which the first two elements are vectorized metric values in 'x', 'y' and 'z' directions for a given pair of subsamples, and the last one is a list of normalized distances between these vectors.        
        """
        return self.vectorizer.vectorize(v1, v2)


class C2(BasicCFMetric):
    """
    Class describing metric C2. 
    """ 
    def __init__(self,  vectorizer, n_threads = 1, show_time=False, normalize=True):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation, default: 1;
        	
        	show_time (bool): flag to monitor time cost for large images, default: False;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. Default: True.
        """ 
        if not isinstance(vectorizer, CFVectorizer):
            raise TypeError("Vectorizer should be an object of CFVectorizer class")
        super().__init__(vectorizer, n_threads, show_time, normalize)
        self.directional = True
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the correlation function C2 for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        return super().generate(cut, cut_name, outputdir, method = 'c2')

    def show(self, inputdir, step, cut_id):
        """
        Vizualize the correlation function C2 for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index.
        """
        title = self.__class__.__name__ + ", " + "step = " + str(step) + ", id = " + str(cut_id)
        super().show(inputdir, step, cut_id, title, 'directional')


class L2(BasicCFMetric):
    """
    Class describing metric L2. 
    """ 
    def __init__(self,  vectorizer, n_threads = 1, show_time=False, normalize=True):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation, default: 1;
        	
        	show_time (bool): flag to monitor time cost for large images, default: False;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. Default: True.
        """
        if not isinstance(vectorizer, CFVectorizer):
            raise TypeError("Vectorizer should be an object of CFVectorizer class")
        super().__init__(vectorizer, n_threads, show_time, normalize)
        self.directional = True
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the correlation function L2 for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        return super().generate(cut, cut_name, outputdir, method = 'l2')

    def show(self, inputdir, step, cut_id):
        """
        Vizualize the correlation function L2 for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index.
        """
        title = self.__class__.__name__ + ", " + "step = " + str(step) + ", id = " + str(cut_id)
        super().show(inputdir, step, cut_id, title, 'directional')


class S2(BasicCFMetric): 
    """
    Class describing metric S2. 
    """ 
    def __init__(self,  vectorizer, n_threads = 1, show_time=False, normalize=True):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation, default: 1;
        	
        	show_time (bool): flag to monitor time cost for large images, default: False;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. Default: True.
        """
        if not isinstance(vectorizer, CFVectorizer):
            raise TypeError("Vectorizer should be an object of CFVectorizer class")
        super().__init__(vectorizer, n_threads, show_time, normalize)
        self.directional = True
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the correlation function S2 for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        return super().generate(cut, cut_name, outputdir, method = 's2')

    def show(self, inputdir, step, cut_id):
        """
        Vizualize the correlation function S2 for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index.
        """
        title = self.__class__.__name__ + ", " + "step = " + str(step) + ", id = " + str(cut_id)
        super().show(inputdir, step, cut_id, title, 'directional')


class SS(BasicCFMetric):
    """
    Class describing metric SS. 
    """ 
    def __init__(self,  vectorizer, n_threads = 1, show_time=False, normalize=True):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation, default: 1;
        	
        	show_time (bool): flag to monitor time cost for large images, default: False;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. Default: True.
        """
        if not isinstance(vectorizer, CFVectorizer):
            raise TypeError("Vectorizer should be an object of CFVectorizer class")
        super().__init__(vectorizer, n_threads, show_time, normalize)
        self.directional = True
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the correlation function SS for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        return super().generate(cut, cut_name, outputdir, method = 'ss')

    def show(self, inputdir, step, cut_id):
        """
        Vizualize the correlation function SS for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index.
        """
        title = self.__class__.__name__ + ", " + "step = " + str(step) + ", id = " + str(cut_id)
        super().show(inputdir, step, cut_id, title, 'directional')


class SV(BasicCFMetric):
    """
    Class describing metric SV. 
    """  
    def __init__(self,  vectorizer, n_threads = 1, show_time=False, normalize=True):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation, default: 1;
        	
        	show_time (bool): flag to monitor time cost for large images, default: False;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. Default: True.
        """ 
        if not isinstance(vectorizer, CFVectorizer):
            raise TypeError("Vectorizer should be an object of CFVectorizer class")
        super().__init__(vectorizer, n_threads, show_time, normalize)
        self.directional = True
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the correlation function SV for a specific subsample.
        
               **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        return super().generate(cut, cut_name, outputdir, method = 'sv')

    def show(self, inputdir, step, cut_id):
        """
        Vizualize the correlation function SV for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index.
        """
        title = self.__class__.__name__ + ", " + "step = " + str(step) + ", id = " + str(cut_id)
        super().show(inputdir, step, cut_id, title, 'directional')


class ChordLength(BasicCFMetric):
    """
    Class describing metric chord length. 
    """
    def __init__(self,  vectorizer, n_threads = 1, show_time=False, normalize=True):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation, default: 1;
        	
        	show_time (bool): flag to monitor time cost for large images, default: False;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. Default: True.
        """
        if not isinstance(vectorizer, DirectVectorizer):
            raise TypeError("Vectorizer should be an object of HistVectorizer class")
        super().__init__(vectorizer, n_threads, show_time, normalize)
        self.directional = False
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the correlation function chord-length for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        return super().generate(cut, cut_name, outputdir, method = 'cl')

    def show(self, inputdir, step, cut_id):
        """
        Vizualize the correlation function chord-length for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index.
        """
        title = self.__class__.__name__ + ", " + "step = " + str(step) + ", id = " + str(cut_id)
        super().show(inputdir, step, cut_id, title, 'density')

class PoreSize(BasicCFMetric):
    """
    Class describing metric pore size. 
    """
    def __init__(self,  vectorizer, n_threads = 1, show_time=False, normalize=True):
        """
        **Input:**
        
        	vectorizer (CFVectorizer): vectorizer to be used for CF metric;
            
            n_threads (int): number of threads used for data generation, default: 1;
        	
        	show_time (bool): flag to monitor time cost for large images, default: False;
        	
        	normalize (bool): flag to control normalization of CF. If True, CF are normalized to satisfy the condition CF(0) = 1. See the details in Karsanina et al. (2021). Compressing soil structural information into parameterized correlation functions. European Journal of Soil Science, 72(2), 561-577. Default: True.
        """
        if not isinstance(vectorizer, DirectVectorizer):
            raise TypeError("Vectorizer should be an object of HistVectorizer class")
        super().__init__(vectorizer, n_threads, show_time, normalize)
        self.directional = False
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the correlation function pore-size for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder.
        """
        return super().generate(cut, cut_name, outputdir, method = 'ps')

    def show(self, inputdir, step, cut_id):
        """
        Vizualize the correlation function chord-length for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index.
        """
        title = self.__class__.__name__ + ", " + "step = " + str(step) + ", id = " + str(cut_id)
        super().show(inputdir, step, cut_id, title, 'density')


# -*- coding: utf-8 -*-
"""Main module of the library, ensures the complete pipeline for REV analysis.
"""
import numpy as np
import os
import json
import shutil
import matplotlib.pyplot as plt
import itertools
import multiprocessing
from functools import partial
from statistics import geometric_mean
from .generators import run_fdmss, _read_array, _write_array, _subcube_ids, make_cut, generate_PNM
from .metrics import BasicMetric, BasicPNMMetric, BasicPDMetric, Permeability
from .REV_formulas import _delta, get_sREV_size, get_dREV_size_1_scalar, get_dREV_size_2_scalar, get_dREV_size_1_vector, get_dREV_size_1_scalar_dimensional, get_dREV_size_2_scalar_dimensional


class REVAnalyzer:
    """
    analysis of representativity of a given image for a given scalar or vector metric.
    """
    def __init__(self, metric, image, size, n_steps, sREV_max_step, datadir=None, outputdir='output'):
        """
        **Input:**

        	metric (subclass of BasicMetric): metric to be analyzed;
        
        	image (str or numpy.ndarray): name of binary ('uint8') file or numpy.ndarray representing the image;
            
            size (tuple (int, int, int)): linear image sizes in x, y and z directions;
                
        	n_steps (int): number of subsamples selection steps;
        
        	sREV_max_step (int): maximal step for which sREV and stationarity analysis can be performed;
        
        	datadir (str): path to the folder containing image, default: None;
        
        	outputdir (str): path to the output folder containing generated data, default: 'output'.
        """
        if not issubclass(metric.__class__, BasicMetric):
            raise TypeError("Metric should be an object of a class derived from BasicMetric.")
        self.metric = metric
        self.size = size
        self._outputdirs_cut_values = []            
        if isinstance(image, str):
            self._outputdir_cut_values = os.path.join(outputdir, image, self.metric.__class__.__name__, 'cuts_values')
            if issubclass(self.metric.__class__, BasicPDMetric):
                for i in range(3):
                    outputdir_cut_values = os.path.join(outputdir, image, 'PD'+str(i), 'cuts_values')
                    self._outputdirs_cut_values.append(outputdir_cut_values)
                    os.makedirs(outputdir_cut_values, exist_ok=True)
        elif isinstance(image, np.ndarray):
            self._outputdir_cut_values = os.path.join(outputdir, self.metric.__class__.__name__, 'cuts_values')
            if issubclass(self.metric.__class__, BasicPDMetric):
                for i in range(3):
                    outputdir_cut_values = os.path.join(outputdir, 'PD'+str(i), 'cuts_values')
                    self._outputdirs_cut_values.append(outputdir_cut_values)
                    os.makedirs(outputdir_cut_values, exist_ok=True)
        else:
            raise TypeError("Image type should be str or numpy.ndarray")
        self.image = image
        self.n_steps = n_steps
        self.sREV_max_step = sREV_max_step
        self.cut_step = (np.array(size)/n_steps).astype(int)
        self.datadir = datadir
        self.outputdir = outputdir
        self.gendatadir = None
        if isinstance(self.image, str):
            self._outputdir_vectorized_cut_values = os.path.join(self.outputdir, self.image, self.metric.__class__.__name__, 'vectorized_cuts_values')
        else:
            self._outputdir_vectorized_cut_values = os.path.join(self.outputdir, self.metric.__class__.__name__, 'vectorized_cuts_values')
        os.makedirs(self.outputdir, exist_ok=True)
        os.makedirs(self._outputdir_cut_values, exist_ok=True)
        self.cut_sizes = [(self.cut_step*(i+1)).tolist() for i in range(self.n_steps-1)]
        if self.size[0] == self.size[1] and self.size[0] == self.size[2]:
            self.geom_mean_cut_sizes = [size[0] for size in self.cut_sizes]
            self.geom_mean_cut_sizes.append(self.size[0])
        else:
            self.geom_mean_cut_sizes = [int(geometric_mean(size)) for size in self.cut_sizes]
            self.geom_mean_cut_sizes.append(int(geometric_mean(self.size)))
        self._sizes_dict = dict(zip(np.arange(1, n_steps+1), self.geom_mean_cut_sizes))
        self._sizes_dict[None] = None
        self.cut_ids = _subcube_ids(self.n_steps, self.sREV_max_step)
        self._metric_cut_names = []
        self.metric_mean = {}
        self.metric_std = {}
        self.metric_normed_std = {}
        self.metric_normed_std_1 = {}
        self.metric_normed_std_2 = {}
        self.dREV_threshold = None
        self.sREV_threshold = None
        self.stationarity_threshold = None
        self.sREV_size_1 = None
        self.sREV_size_2 = None
        self.dREV_size_1 = None
        self.dREV_size_2 = None
        self.is_fdmss_data = False
        self.is_pnm_data = False
        if isinstance(self.metric, Permeability):
            if not (self.size[0] == self.size[1] and self.size[0] == self.size[2]):
                raise ValueError("Only cubical images can be input for FDMSS solver.")
            if isinstance(self.image, str):
                self.gendatadir = os.path.join(self.outputdir, self.image, 'fdmss_data')
            else:
                self.gendatadir = os.path.join(self.outputdir, 'fdmss_data')
            fdmss_input=os.path.join(self.gendatadir, 'fdmss_input.txt')
            if os.path.isfile(fdmss_input):
                with open(fdmss_input) as f:
                    lines = [line.rstrip('\n') for line in f]
                    if lines[0] == self.metric.direction and lines[1] == str(self.metric.resolution):                        
                        self.is_fdmss_data = True
        if issubclass(self.metric.__class__, BasicPNMMetric):
            if isinstance(self.image, str):
                self.gendatadir = os.path.join(self.outputdir, self.image, 'pn_data')
            else:
                self.gendatadir = os.path.join(self.outputdir, 'pn_data')
            pn_input=os.path.join(self.gendatadir, 'pn_input.txt')
            if os.path.isfile(pn_input):
                with open(pn_input) as f:
                    lines = [line.rstrip('\n') for line in f]
                if lines[0] == str(self.n_steps) and lines[1] == str(self.sREV_max_step) and lines[2] == str(self.metric.resolution): 
                    self.is_pnm_data = True
                    

    def generate(self):
        """
        Generator of metric values for all selected subsamples.
        """
        if isinstance(self.metric, Permeability) and not self.is_fdmss_data:
            os.makedirs(self.gendatadir, exist_ok=True)
            if isinstance(self.image, str):
                run_fdmss(self.image, self.metric.direction, self.datadir, self.gendatadir, self.metric.n_threads, self.metric.resolution, self.metric.show_time)
            else:
                fileout = os.path.join(self.gendatadir, 'image.raw') 
                _write_array(self.image, fileout)
                run_fdmss('image.raw', self.metric.direction, self.gendatadir, self.gendatadir, self.metric.n_threads, self.metric.resolution, self.metric.show_time)
                os.remove(fileout)
            fdmss_input = os.path.join(self.gendatadir, 'fdmss_input.txt')            
            with open(fdmss_input, 'w') as f:
                lines = [self.metric.direction, str(self.metric.resolution), str(self.size[0])]
                f.write('\n'.join(lines))
        if isinstance(self.image, str):
            if self.datadir is not None:
                filein = os.path.join(self.datadir, self.image)
            else:
                filein = self.image
            image = _read_array(filein, self.size[0], self.size[1], self.size[2], 'uint8')
        else:
            image = self.image
        if issubclass(self.metric.__class__, BasicPNMMetric) and not self.is_pnm_data:
            os.makedirs(self.gendatadir, exist_ok=True)
            generate_PNM(image, self.size, self.n_steps, self.sREV_max_step, self.gendatadir, self.metric.n_threads, self.metric.resolution, self.metric.show_time)     
            pn_input = os.path.join(self.gendatadir, 'pn_input.txt')            
            with open(pn_input, 'w') as f:
                lines = [str(self.n_steps), str(self.sREV_max_step), str(self.metric.resolution)]
                f.write('\n'.join(lines))
        ids = _subcube_ids(self.n_steps, self.sREV_max_step)
        cuts = []
        for elem in ids:
            l = elem[0]
            idx = elem[1]
            if l  == self.n_steps:
                cuts.append(image)
            else:
                cut_size = self.cut_sizes[l-1]
                cuts.append(make_cut(image, self.size, cut_size, idx))
        data = zip(ids, cuts)
        pool = multiprocessing.Pool(processes=self.metric.n_threads) 
        results = pool.map(self._metric_for_subsample, data)
        pool.close()
        pool.join()           

    def read(self, step, cut_id=0): 
        """
        Read the generated metric value for a given subsample.
        
        **Input:**
        
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index .   
        
        **Output**
        
        	metric value (float or np.array(dtype='float')).       
        """
        return self.metric.read(
            self._outputdir_cut_values, step, cut_id)
    
    def show(self, step, cut_id = 0, nbins = None):
        """
        Vizualize the vector metric for a specific subsample.
        
        **Input:**
         
        	step (int): subsamples selection step;
        
        	cut_id (int: 0,..8): cut index;
        	
        	nbins (int): number of bins in histogram. For PNM-based metric only.
        """
        if not self.metric.metric_type == 'v':
            raise TypeError("Metric type should be vector")
        if issubclass(self.metric.__class__, BasicPNMMetric):
            if nbins is None:
                raise ValueError("Number of bins in histogram should be defined for the visualization of this metric")
            self.metric.show(self._outputdir_cut_values, step, cut_id, nbins)
        else:
            self.metric.show(self._outputdir_cut_values, step, cut_id)
        

    def vectorize(self):
        """
        Vectorization of generated metric data using vetorizer. For vector metric only.
        """
        if not self.metric.metric_type == 'v':
            raise TypeError("Metric type should be vector")
        os.makedirs(self._outputdir_vectorized_cut_values, exist_ok=True)
        x = np.arange(9)
        y = [0]
        steps = [i for i in range(1, self.n_steps)]
        data1 = []
        data2 = []
        data3 = []
        for step in steps:
            if step > self.sREV_max_step:
                data1 = data1 + [tup for tup in itertools.product([step], y, y)]
            elif step == self.sREV_max_step:
                data2 = data2 + [tup for tup in itertools.product([step], x, y)]
            else:
                data3 = data3 + [tup for tup in itertools.product([step], x, x)]
        data = data1 + data2 + data3
        pool = multiprocessing.Pool(processes=self.metric.n_threads) 
        results = pool.map(self._vectorize_subsample, data)
        pool.close()
        pool.join() 
        ds = [{} for i in range(1, self.n_steps)]
        for elem in results:
            ds[elem[0]-1][elem[1]] = elem[2]      
        for step in steps:     
            jsonname = 'cut_' + str(step)
            with open(os.path.join(self._outputdir_vectorized_cut_values, jsonname), 'w') as f:
                json.dump(ds[step-1], f, indent=4)     
            

    def analyze(self, dREV_threshold, sREV_threshold):
        """
        Perform the analysis of representativity.
        
        **Input:**
        
        	dREV_threshold (float, <1): threshold to estimate dREV size;
        	
        	sREV_threshold (float, <1): threshold to estimate sREV size.
        """
        self.dREV_threshold = dREV_threshold
        self.sREV_threshold = sREV_threshold
        if self.metric.metric_type == 's':
            for l in range(1, self.n_steps+1):
                if (l <= self.sREV_max_step):
                    data = [self.read(l, i) for i in range(9)]
                    if self.metric.directional:
                        self.metric_mean[l] = [
                            np.mean(list(i)) for i in zip(*data)]
                        self.metric_std[l] = [
                            np.std(list(i)) for i in zip(*data)]
                        self.metric_normed_std[l] = max(
                            [np.std(list(i))/abs(np.mean(list(i))) for i in zip(*data)])
                    else:
                        self.metric_mean[l] = np.mean(data)
                        self.metric_std[l] = np.std(data)
                        self.metric_normed_std[l] = self.metric_std[l] / \
                            abs(self.metric_mean[l])
                else:
                    if self.metric.directional:
                        self.metric_mean[l] = self.read(l, 0)
                    else:
                        self.metric_mean[l] = self.read(l, 0).item()
            self.sREV_size_1 = self._sizes_dict[get_sREV_size(
                self.metric_normed_std, self.sREV_threshold)]
            if self.metric.directional:
                self.dREV_size_1 = self._sizes_dict[get_dREV_size_1_scalar_dimensional(
                    self.metric_mean, self.dREV_threshold)]
                self.dREV_size_2 = self._sizes_dict[get_dREV_size_2_scalar_dimensional(
                    self.metric_mean, self.dREV_threshold)]
            else:
                self.dREV_size_1 = self._sizes_dict[get_dREV_size_1_scalar(
                    self.metric_mean, self.dREV_threshold)]
                self.dREV_size_2 = self._sizes_dict[get_dREV_size_2_scalar(
                    self.metric_mean, self.dREV_threshold)]
        if self.metric.metric_type == 'v':
            for l in range(1, self.n_steps):
                jsonname = 'cut_' + str(l)
                with open(os.path.join(self._outputdir_vectorized_cut_values, jsonname), 'r') as f:
                    d = json.load(f)
                deltas = [elem for elem in d.values()]
                n = deltas[0]
                if (type(n) is list):
                    self.metric_mean[l] = max(
                        [np.mean(list(i)) for i in zip(*deltas)])
                else:
                    self.metric_mean[l] = np.mean(deltas)
                if (l <= self.sREV_max_step):
                    if (type(n) is list):
                        self.metric_std[l] = max(
                            [np.std(list(i)) for i in zip(*deltas)])
                        self.metric_normed_std[l] = max(
                            [np.std(list(i))/np.mean(list(i)) for i in zip(*deltas)])
                    else:
                        self.metric_std[l] = np.std(deltas)
                        self.metric_normed_std[l] = self.metric_std[l] / \
                            self.metric_mean[l]
                    self.metric_normed_std_1[l] = self.metric_std[l] / \
                        self.dREV_threshold
                    if (self.dREV_threshold is None or self.metric_mean[l] > self.dREV_threshold):
                        self.metric_normed_std_2[l] = self.metric_normed_std[l]
                    else:
                        self.metric_normed_std_2[l] = self.metric_normed_std_1[l]
            self.dREV_size_1 = self._sizes_dict[get_dREV_size_1_vector(
                self.metric_mean, self.dREV_threshold)]
            self.sREV_size_1 = self._sizes_dict[get_sREV_size(
                self.metric_normed_std_1, self.sREV_threshold)]
            self.sREV_size_2 = self._sizes_dict[get_sREV_size(
                self.metric_normed_std_2, self.sREV_threshold)]

    def analyze_stationarity(self, stationarity_threshold):
        """
        Perform the analysis of stationarity.
        
        **Input:**
        
        	stationarity_threshold (float, <1): threshold to analyze stationarity.
        	
        **Output**
        
        	 True or False: is image stationary.
        """
        if not self.metric.metric_type == 'v':
            raise TypeError("Metric type should be vector")
        self.stationarity_threshold = stationarity_threshold
        x = np.arange(9)
        for step in range(1, self.sREV_max_step):
            ids = itertools.combinations_with_replacement(x, 2)
            pool = multiprocessing.Pool(processes=self.metric.n_threads) 
            results = pool.map(partial(self._distance_for_subsamples, step=step), ids)
            pool.close()
            pool.join()
            dmax = max(results)
            print("at step ", step, " maximal distance between subsamples is ", dmax)
            if dmax > self.stationarity_threshold:
                print("Image is nonstationary.")
                return False
        print("Image is stationary.")
        return True
            
    
    def show_results(self):
        """
        Visualization of REV analysis results.
        """
        plt.rcParams.update({'font.size': 16})
        plt.rcParams['figure.dpi'] = 300
        x = list(self.metric_mean.keys())
        x.sort()
        xerr = list(self.metric_std.keys())
        xerr.sort()
        nzeros = len(x) - len(xerr)
        fig, ax = plt.subplots()
        title = self.metric.__class__.__name__ 
        plt.title(title)
        if self.metric.metric_type == 's' and self.metric.directional:
            y = [self.metric_mean[l] for l in x]
            y1 = [i for i in zip(*y)]
            yerr = [self.metric_std[l] for l in xerr]
            for i in range(nzeros):
                yerr.append([0, 0, 0])
            yerr1 = [i for i in zip(*yerr)]
            x0 = [self._sizes_dict[i] for i in x]
            x1 = np.array(x0)-2
            x2 = np.array(x0)+2
            plt.errorbar(x1, y1[0], yerr=yerr1[0], label='x')
            plt.errorbar(x0, y1[1], yerr=yerr1[1], label='y')
            plt.errorbar(x2, y1[2], yerr=yerr1[2], label='z')
            plt.legend()
        else:
            y = [self.metric_mean[l] for l in x]
            yerr = [self.metric_std[l] for l in xerr]
            for i in range(nzeros):
                yerr.append(0)
            x1 = [self._sizes_dict[i] for i in x]
            plt.errorbar(x1, y, yerr=yerr)
        plt.xlabel("Mean size of subsample")
        if self.metric.metric_type == 's':
            plt.ylabel(self.metric.__class__.__name__)
        if self.metric.metric_type == 'v':
            plt.ylabel("difference in distance")
            ax.axhline(y=self.dREV_threshold, color='k',
                       linestyle='-', label="dREV threshold")
        plt.show()

        fig, ax = plt.subplots()
        plt.title(title)
        xdrev = x[:-1]
        xsrev = list(self.metric_normed_std.keys())
        if self.metric.metric_type == 's':
            if self.metric.directional:
                y1 = np.array(y1)
                ydrev1 = [max([_delta(y1[k][i], y1[k][i+1])
                              for k in range(3)]) for i in range(len(xdrev))]
                ydrev2 = [max([_delta(y1[k][i], y1[k][-1])
                              for k in range(3)]) for i in range(len(xdrev))]
            else:
                ydrev1 = [_delta(self.metric_mean[x[i]], self.metric_mean[x[i+1]])
                          for i in range(len(xdrev))]
                y0 = self.metric_mean[x[-1]]
                ydrev2 = [_delta(self.metric_mean[x[i]], y0)
                          for i in range(len(xdrev))]
            ysrev = [self.metric_normed_std[l] for l in xsrev]
            xsrev = [self._sizes_dict[i] for i in xsrev]
            xdrev = [self._sizes_dict[i] for i in xdrev]
            ax.plot(xsrev, ysrev, "r--", label='sREV, $\sigma_{norm}$')
            ax.plot(xdrev, ydrev1, "b-", label='dREV, $\delta_1$')
            ax.plot(xdrev, ydrev2, "g-", label='dREV, $\delta_2$')
            ax.axhline(y=self.dREV_threshold, color='k',
                       linestyle='-', label="dREV threshold")
            plt.ylabel('$\sigma_{norm}$, $\delta_1$, $\delta_2$')
        if self.metric.metric_type == 'v':
            ysrev1 = [self.metric_normed_std_1[l] for l in xsrev]
            ysrev2 = [self.metric_normed_std_2[l] for l in xsrev]
            xsrev = [self._sizes_dict[i] for i in xsrev]
            ax.plot(xsrev, ysrev1, "r-", label='sREV, $\sigma_{norm1}$')
            ax.plot(xsrev, ysrev2, "r--", label='sREV, $\sigma_{norm2}$')
            plt.ylabel("$\sigma_{norm1}$, $\sigma_{norm2}$")
        ax.axhline(y=self.sREV_threshold, color='k',
                   linestyle='--', label="sREV threshold")
        plt.xlabel("Mean size of subsample")

        plt.legend()
        plt.show()
    
    def _metric_for_subsample(self, data):
        if issubclass(self.metric.__class__, BasicPDMetric):
            outputdir = self._outputdirs_cut_values
        else:
            outputdir = self._outputdir_cut_values
        l = data[0][0]
        idx = data[0][1]
        cut = data[1]
        cut_name = 'cut'+str(l)+'_'+str(idx)
        result = self.metric.generate(cut, cut_name, outputdir, self.gendatadir)
    
    def _vectorize_subsample(self, data):
        step = data[0]
        v1 = self.read(step, data[1])
        if self.metric.directional:
            v0 = v1[0]
        else:
            v0 = v1
        if len(v0) == 0:
            return np.nan
        v2 = self.read(step + 1, data[2])
        str_elem = str(data[1]) + ', ' + str(data[2])
        result = self.metric.vectorize(v1, v2)
        if (type(result[2]) is list and (np.nan in result[2])) or (type(result[2]) is not list and np.isnan(result[2])):
            return np,nan
        return (step, str_elem, result[2])
     
    def _distance_for_subsamples(self, ids, step):
        v1 = self.read(step, ids[0])
        if ids[0] == ids[1]:
            v2 = self.read(step + 1, ids[0])
        else:
            v2 = self.read(step, ids[1])
        result = self.metric.vectorize(v1, v2)
        delta = result[2]
        if (type(delta) is list):
            return max(delta)
        else:
            return delta
        



# -*- coding: utf-8 -*-
"""Module for running FDMSS solver."""

import os
import numpy as np
import time
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ElementTree
from .utils import _read_array, make_cut
from pyfdmss import run

fdmss_data = "fdmss_data"


def run_fdmss(image, direction, datadir, outputdir, n_threads=1, resolution=1., show_time=False):
    """
    Running FDMSS solver for an initial image.
    
    **Input:**

     	image (str): name of binary ('uint8') file representing the image;     
     	
     	direction (str): 'x', 'y', 'z' or 'all'. If label of this parameter is 'all', all 3 possible flow directions are considered;
     	
        datadir (str): path to the folder containing image;
     	
     	outputdir (str): path to the output folder containing generated data;
        
     	n_threads (int): number of threads used by FDMSS, default: 1;
     	
     	resolution (float): resolution of studied sample (micrometers), default: 1;
     	
     	show_time (bool): Added to monitor time cost for large images,  default: False. 
    """
    if not (direction == 'x' or direction == 'y' or direction == 'z' or direction == 'all'):
        raise ValueError("Direction should be 'x', 'y', 'z' or 'all'")
    if direction == 'all':
        directions_list = ['x', 'y', 'z']
    else:
        directions_list = [direction]
    for d in directions_list:
        start_time = time.time()
        if datadir is not None:
            image_path = os.path.join(datadir, image)
        else:
            image_path = image
        config_path = os.path.join(outputdir, d + '_config.xml')
        _make_fdmss_config(config_path, d, resolution, n_threads)
        summary_path = os.path.join(outputdir, d + '_summary.xml')
        pressure_path = os.path.join(outputdir, d + '_pressure')
        velx_path = os.path.join(outputdir, d + '_velx')
        vely_path = os.path.join(outputdir, d + '_vely')
        velz_path = os.path.join(outputdir, d + '_velz')
        full_vel_path =  os.path.join(outputdir, d + '_full_vel')
        comp_vel_path =  os.path.join(outputdir, d + '_comp_vel')
        log_path = 'log.txt'
        run(config_path, image_path, summary_path, velx_path, vely_path, velz_path, pressure_path, full_vel_path, comp_vel_path, log_path)
        if show_time:
            print("---fdmss run time is %s seconds ---" % (time.time() - start_time))


def _make_fdmss_config(fileout, direction, resolution=1, n_threads=1):
    xml_data = """
<OverdozenPermsolverParams>
  <Parameter name="Resolution" datatype="float">
    <ParameterValue>1.0</ParameterValue>
  </Parameter>
  <Parameter name="TimeStep" datatype="float">
    <ParameterValue>0.0003</ParameterValue>
  </Parameter>
  <Parameter name="IterationsPerStep" datatype="unsigned int">
    <ParameterValue>100</ParameterValue>
  </Parameter>
  <Parameter name="MaximumStepsCount" datatype="unsigned int">
    <ParameterValue>140</ParameterValue>
  </Parameter>
  <Parameter name="InitialVelocityValue" datatype="float">
    <ParameterValue>0.001</ParameterValue>
  </Parameter>
  <Parameter name="Accuracy" datatype="float">
    <ParameterValue>0.01</ParameterValue>
  </Parameter>
  <Parameter name="BoundaryCondition" datatype="int">
    <ParameterValue>1</ParameterValue>
  </Parameter>
  <Parameter name="TerminationCondition" datatype="int">
    <ParameterValue>2</ParameterValue>
  </Parameter>
  <Parameter name="AccuracyOrder" datatype="int">
    <ParameterValue>0</ParameterValue>
  </Parameter>
  <Parameter name="ErrorSmoothingLength" datatype="unsigned int">
    <ParameterValue>10</ParameterValue>
  </Parameter>
  <Parameter name="ThreadsNumber" datatype="unsigned int">
    <ParameterValue>1</ParameterValue>
  </Parameter>
  <Parameter name="FlowDirectionAxis" datatype="char">
    <ParameterValue>AXIS</ParameterValue>
  </Parameter>
  <Parameter name="WaterLayerWidth" datatype="unsigned int">
    <ParameterValue>0</ParameterValue>
  </Parameter>
</OverdozenPermsolverParams>
"""
    root = ET.fromstring(xml_data)
    root[0][0].text = str(resolution)
    root[10][0].text = str(n_threads)
    root[11][0].text = direction
    tree = ElementTree(root)
    tree.write(fileout)


def _pressure_diff(image, pressure, axis):
    inv = ~image
    pressure = np.where(inv, pressure, 0)
    if axis == 'x':
        p_start = np.sum(pressure[:, :, 1])/np.sum(inv[:, :, 1])
        p_end = np.sum(pressure[:, :, -1])/np.sum(inv[:, :, -1])
    if axis == 'y':
        p_start = np.sum(pressure[:, 1, :])/np.sum(inv[:, 1, :])
        p_end = np.sum(pressure[:, -1, :])/np.sum(inv[:, -1, :])
    if axis == 'z':
        p_start = np.sum(pressure[1, :, :])/np.sum(inv[1, :, :])
        p_end = np.sum(pressure[-1, :, :])/np.sum(inv[-1, :, :])
    return (p_start - p_end)/image.shape[0]


def _get_permeability(image, porosity, pressure, vel, direction):
    dim = image.shape[0]
    pores = dim**3 - np.count_nonzero(image)
    p = _pressure_diff(image, pressure, direction)
    v = np.sum(vel)/pores
    return 100*v/p*porosity/0.986*1000


def _get_porosity(image):
    dim = image.shape[0]
    pores = dim**3 - np.count_nonzero(image)
    return pores/(dim**3)



# -*- coding: utf-8 -*-
"""
Definition of Persistence Image vectorizer, used in PD vectorization. A persistence image is a 2D function computed from a persistence diagram by convolving the diagram points with a weighted Gaussian kernel. The plane is then discretized into an image with pixels, which is flattened and returned as a vector. See Adams H. et al., (2017). Persistence images: A stable vector representation of persistent homology. Journal of Machine Learning Research, 18.
"""

import numpy as np
from gudhi.representations.vector_methods import PersistenceImage
from .basic_vectorizer import BasicVectorizer


class PersistenceImageVectorizer(BasicVectorizer):
    """
    Class describing persistence image vectorizer.
    """      
    def __init__(self, resolution, bandwidth=1., norm=2):
        """
        **Input:**
        
        	resolution ([int, int]): number of bins at each axe in XY plane in vectorized persistence image function;
        	
        	bandwidth (float): bandwidth of the Gaussian kernel, default: 1;
        	 
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function, default: 2.
        """  
        super().__init__(norm)
        self.resolution = resolution
        self.bandwidth = bandwidth

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        	(list(dtype = float), list(dtype = float), float): a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        """    
        n1 = len(v1)
        n2 = len(v2)
        C1 = PersistenceImage(bandwidth=self.bandwidth, weight=lambda x: np.arctan(0.5*(x[1] - x[0])),
                              resolution=self.resolution)
        C1.fit([v1])
        range1 = C1.im_range
        v1 = C1.transform([v1])
        v1 = v1/n1
        C2 = PersistenceImage(bandwidth=self.bandwidth, weight=lambda x: np.arctan(0.5*(x[1] - x[0])),
                              resolution=self.resolution, im_range=range1)
        C2.fit([v2])
        v2 = C2.transform([v2])
        v2 = v2/n2
        n1 = np.linalg.norm(v1, ord=self.norm)
        n2 = np.linalg.norm(v2, ord=self.norm)
        delta = 2*np.linalg.norm(np.array(v1) -
                                 np.array(v2), ord=self.norm)/(n1 + n2)
        return v1.tolist(), v2.tolist(), delta



# -*- coding: utf-8 -*-
"""Definition of direct vectorizer"""

import numpy as np
from .basic_vectorizer import BasicVectorizer


class DirectVectorizer(BasicVectorizer):
    """
    Class describing direct vectorizer.
    """   
    def __init__(self, norm=2):
        """
        **Input:**
        
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function, default: 2.
        """      
        super().__init__(norm)

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
       		(list(dtype = float), list(dtype = float), float) - a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors.        
        """
        n = min(len(v1), len(v2))
        v1 = v1[:n]
        v2 = v2[:n]
        v_norm1 = np.linalg.norm(v1, ord=self.norm)
        v_norm2 = np.linalg.norm(v2, ord=self.norm)
        deltas = 2 * np.linalg.norm(np.array(v1) - np.array(v2), ord=self.norm)/(v_norm1 + v_norm2)
        return v1.tolist(), v2.tolist(), deltas



# -*- coding: utf-8 -*-
"""Vectorizers for vector metrics."""

from .basic_vectorizer import BasicVectorizer
from .hist_vectorizer import HistVectorizer
from .direct_vectorizer import DirectVectorizer
from .cf_vectorizer import CFVectorizer
from .simple_binning_vectorizer import SimpleBinningVectorizer
from .persistence_image_vectorizer import PersistenceImageVectorizer
from .landscape_vectorizer import LandscapeVectorizer
from .silhouette_vectorizer import SilhouetteVectorizer



# -*- coding: utf-8 -*-
"""Definition of Simple Binning vectorizer, used in PD vectorization."""

import numpy as np
from .basic_vectorizer import BasicVectorizer


class SimpleBinningVectorizer(BasicVectorizer):
    """
    Class describing simple binning vectorizer.
    """  
    def __init__(self, bins, skip_zeros=True, norm=2):
        """
        **Input:**
        
        	bins (int): number of bins at each axe in XY plane;
        
        	skip_zeros (bool): If True, bins of 2D histogram empty for both compared PDs are not included into the final vectors, default: True;
        	
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function, default: 2.
        """         
        super().__init__(norm)
        self.bins = bins
        self.skip_zeros = skip_zeros

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        	(list(dtype = float), list(dtype = float), float): a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        """        
        r1 = _range_pd(v1)
        v1 = _hist_pd(v1, self.bins, r1)
        v2 = _hist_pd(v2, self.bins, r1)
        if self.skip_zeros:
            v12 = _skip_zeros12(v1, v2)
            v1 = v12[0]
            v2 = v12[1]
        n1 = np.linalg.norm(v1, ord=self.norm)
        n2 = np.linalg.norm(v2, ord=self.norm)
        delta = 2*np.linalg.norm(np.array(v1) -
                                 np.array(v2), ord=self.norm)/(n1 + n2)
        return v1, v2, delta


def _hist_pd(v, bins, r):
    v = [i for i in zip(*v)]
    b = np.array(v[0])
    d = np.array(v[1])
    d = d - b
    hist = np.histogram2d(b, d, bins=bins, range=r)
    norm = len(b)
    return np.ravel(hist[0])/norm


def _range_pd(v):
    v = [i for i in zip(*v)]
    b = np.array(v[0])
    d = np.array(v[1])
    d = np.array(d) - np.array(b)
    xmin = min(b)
    xmax = max(b)
    ymin = min(d)
    ymax = max(d)
    r = [[xmin, xmax], [ymin, ymax]]
    return r


def _skip_zeros12(v1, v2):
    v1_new = []
    v2_new = []
    zeroes_ids = []
    for count, elem in enumerate(zip(v1, v2)):
        if (elem[0] > 0 or elem[1] > 0):
            v1_new.append(elem[0])
            v2_new.append(elem[1])
        else:
            zeroes_ids.append(count)
    return (v1_new, v2_new, zeroes_ids)



# -*- coding: utf-8 -*-
"""Definition of Permeability metric."""

import numpy as np
import os
from .basic_metric import BasicMetric
from ..generators import _read_array, make_cut


class Permeability(BasicMetric):
    """
    Class describing permeability metric.
    """
    def __init__(self, direction='all', n_threads=1, resolution=1., show_time = False):
        """
        **Input:**
        
         	direction (str): 'x', 'y', 'z' or 'all'. If label of this parameter is 'all', permeability values are generated for all 3 possible flow directions;
            
            n_threads (int): number of threads used by FDMSS, default: 1;
            
            resolution (float): resolution of studied sample (micrometers), default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False. 
        """
        super().__init__(vectorizer=None, n_threads = n_threads)
        self.metric_type = 's'
        self.direction = direction
        self.resolution = resolution
        self.show_time = show_time
        if direction == 'all':
            self.directional = True
            
    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates permeability for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated fdmss data output.    
        """
        fdmss_input = os.path.join(gendatadir, 'fdmss_input.txt')
        with open(fdmss_input) as f:
            lines = [line.rstrip('\n') for line in f]
            L = int(lines[2])
        if self.directional:
            directions_list = ['x', 'y', 'z']
        else:
            directions_list = [self.direction]
        for direction in directions_list:
            pressure_name = os.path.join(gendatadir, direction + '_pressure')
            vel_name  = os.path.join(gendatadir, direction + '_vel' + direction)
            pressure = _read_array(pressure_name, L, L, L, 'float32')
            vel = _read_array(vel_name, L, L, L, 'float32')
            if self.directional:
                cut_name_out = cut_name + "_" + direction + ".txt"
            else:
                cut_name_out = cut_name + ".txt"
            fileout = os.path.join(outputdir, cut_name_out)
            cut_size = (cut.shape[0], cut.shape[1], cut.shape[2])
            idx = int(cut_name[-1])
            size0 = (L, L, L)
            pressure_cut = make_cut(pressure, size0, cut_size, idx)
            vel_cut = make_cut(vel, size0, cut_size, idx)
            permeability = _get_permeability(cut, pressure_cut, vel_cut, direction)
            with open(fileout, "w") as f:
                if permeability > 0:
                    f.write(str(permeability))
                else:
                    f.write(str(0))


def _pressure_diff(image, pressure, axis):
    inv = ~image
    pressure = np.where(inv, pressure, 0)
    if axis == 'x':
        p_start = np.sum(pressure[:, :, 1])/np.sum(inv[:, :, 1])
        p_end = np.sum(pressure[:, :, -1])/np.sum(inv[:, :, -1])
    if axis == 'y':
        p_start = np.sum(pressure[:, 1, :])/np.sum(inv[:, 1, :])
        p_end = np.sum(pressure[:, -1, :])/np.sum(inv[:, -1, :])
    if axis == 'z':
        p_start = np.sum(pressure[1, :, :])/np.sum(inv[1, :, :])
        p_end = np.sum(pressure[-1, :, :])/np.sum(inv[-1, :, :])
    return (p_start - p_end)/image.shape[0]


def _get_permeability(image, pressure, vel, direction):
    dim = image.shape[0]
    pores = dim**3 - np.count_nonzero(image)
    p = _pressure_diff(image, pressure, direction)
    v = np.sum(vel)/pores
    porosity = pores/(dim**3)
    return v/p*porosity





# -*- coding: utf-8 -*-
"""
Metric generators based on external software.
"""

from .pnm_generator import generate_PNM
from .fdmss_generator import run_fdmss
from .utils import make_cut, _read_array, _write_array, _subcube_ids



# -*- coding: utf-8 -*-
"""Definition of CF vectorizer"""

import numpy as np
from .basic_vectorizer import BasicVectorizer


class CFVectorizer(BasicVectorizer):
    """
    Class describing CF vectorizer.
    """   
    def __init__(self, norm=2, mode='max'):
        """
        **Input:**
        
        	mode (str): can be 'all' or 'max'. If mode = 'all', CF calculated for 'x', 'y' and 'z' directions are concatenated into one vector during vectorization. If mode = 'max', CF calculared for different directions are vectorizes  independently. 
        	Then at the analisys step, maximal differences and deviations over 3 directions are taking for REV sizes calculation. Default: 'max';
        	
        	norm (int): Norm of vectors used in REV analysis. The same, as parameter 'ord' in numpy.linalg.norm function, default: 2.
        """      
        super().__init__(norm)
        self.mode = mode

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        Depends on the chosen mode.
        
        	If mode = 'all':
        
        		(list(dtype = float), list(dtype = float), float) - a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        
        	If mode = 'max:
        
        		(list(list(dtype = float)), list(list(dtype = float)), list(float)) - a tuple, in which the first two elements are vectorized metric values in 'x', 'y' and 'z' directions for a given pair of subsamples, and the last one is a list of normalized distances between these vectors.        
        """
        if not (self.mode == 'max' or self.mode == 'all'):
            raise ValueError("Mode should be 'max' or 'all'.")
        n = min(len(v1[0]), len(v2[0]))
        if self.mode == 'max':
            v_res1 = []
            v_res2 = []
            deltas = []
            for i in range(3):
                v_norm1 = np.linalg.norm(v1[i][:n], ord=self.norm)
                v_norm2 = np.linalg.norm(v2[i][:n], ord=self.norm)
                d = np.linalg.norm(np.array(v1[i][:n]) -
                                   np.array(v2[i][:n]), ord=self.norm)
                deltas.append(2 * d/(v_norm1 + v_norm2))
                vi1 = np.array(v1[i][:n])/v_norm1
                vi2 = np.array(v2[i][:n])/v_norm2
                v_res1.append(vi1.tolist())
                v_res2.append(vi2.tolist())
        if self.mode == 'all':
            v_res1 = np.concatenate([v1[0][:n], v1[1][:n], v1[2][:n]]).tolist()
            v_res2 = np.concatenate([v2[0][:n], v2[1][:n], v2[2][:n]]).tolist()
            v_norm1 = np.linalg.norm(np.array(v_res1), ord=self.norm)
            v_norm2 = np.linalg.norm(np.array(v_res2), ord=self.norm)
            deltas = 2 * \
                np.linalg.norm(np.array(v_res1) - np.array(v_res2),
                               ord=self.norm)/(v_norm1 + v_norm2)
            v_res1 = (v_res1/v_norm1).tolist()
            v_res2 = (v_res2/v_norm2).tolist()
        return v_res1, v_res2, deltas



# -*- coding: utf-8 -*-
"""Definition of PNM-based metrics."""

import numpy as np
import matplotlib.pyplot as plt
import os
import pandas as pd
from .basic_metric import BasicMetric
from revanalyzer.vectorizers import HistVectorizer


class BasicPNMMetric(BasicMetric):
    """
    Base class of PNM-based metrics. (Don't use it directly but derive from it).
    """  
    def __init__(self, vectorizer, n_threads, resolution, show_time):
        """
        **Input:**
        
        	vectorizer (PNMVectorizer object): vectorizer to be used for a vector metric.
                        
            n_threads (int): number threads used for data generation;
                    
            resolution (float): resolution of studied sample;
            
            show_time (bool): Added to monitor time cost for large images.
        """
        if not (isinstance(vectorizer, HistVectorizer) or (vectorizer is None)):
            raise TypeError("Vectorizer should be None or an object of HistVectorizer class.")
        super().__init__(vectorizer, n_threads = n_threads)
        self.resolution = resolution
        self.show_time = show_time

    def generate(self, cut_name, gendatadir):
        """
        Generates PNM metric for a specific subsample.
        
        **Input:**
        
            cut_name (str): name of subsample;
        	
        	gendatadir (str): folder with generated PNM data.
            
            **Output:**
        
        	df (pandas.DataFrame): data frame with pnm statistics.
        """
        cut_name = os.path.join(gendatadir, cut_name + '.csv')
        df =  pd.read_csv(cut_name)
        return df
        
        
    def show(self, inputdir, step, cut_id, nbins, metric_name):
        """
        Vizualize the vector metric for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subcubes;
        	
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index;
        	
        	nbins (int): number of bins in histogram;
            
            metric_name(str): name of metric.
        """
        data = self.read(inputdir, step, cut_id)
        data = data/self.resolution 
        max_value = max(data)
        range_data = [0, max_value]
        hist, bin_edges = np.histogram(
            data, bins=nbins, range=range_data, density=True)
        step1 = max_value/nbins
        x = [i*step1 for i in range(nbins)]
        plt.rcParams.update({'font.size': 16})
        plt.rcParams['figure.dpi'] = 300
        fig, ax = plt.subplots(figsize=(10, 8))
        title = metric_name + ", "  + "cut size = " + str(step) + ", id = " + str(cut_id)
        ax.set_title(title)
        ax.bar(x, hist, width=0.5, color='r')
        ax.set_xlabel(metric_name)
        ax.set_ylabel('density')
        plt.show()

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsample. Makes normalization to voxels and calls the
        vectorizer function.
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        	(list(dtype = float), list(dtype = float), float) - a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        """
        if not self.metric_type == 'v':
            raise TypeError("Metric type should be vector")
        v1 = v1/self.resolution
        v2 = v2/self.resolution
        res = self.vectorizer.vectorize(v1, v2)
        return res

class PoreNumber(BasicPNMMetric):
    """
    Class describing pore number metric.
    """ 
    def __init__(self, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False.             
        """
        super().__init__(vectorizer=None, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 's'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates pore number for a specific subsample.
        
        **Input:**
        
            cut (numpy.ndarray): 3D array representing a subsample;
        
            cut_name (str): name of subsample;
            
            outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.  
        """    
        df = super().generate(cut_name, gendatadir)
        dimx = cut.shape[0]
        dimy = cut.shape[1]
        dimz = cut.shape[2]
        volume = dimx*dimy*dimz
        pore_number = (df.shape[0] - df['pore.phase'].isna().sum())/volume
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        with open(fileout, "w") as f:
            f.write(str(pore_number))

class ThroatNumber(BasicPNMMetric):
    """
    Class describing throat number metric.
    """ 
    def __init__(self, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False.             
        """
        super().__init__(vectorizer=None, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 's'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates throat number for a specific subsample.
        
        **Input:**
        
            cut (numpy.ndarray): 3D array representing a subsample;
        
            cut_name (str): name of subsample;
            
            outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.  
        """
        dimx = cut.shape[0]
        dimy = cut.shape[1]
        dimz = cut.shape[2]
        volume = dimx*dimy*dimz
        df = super().generate(cut_name, gendatadir)
        throat_number = (df.shape[0] - df['throat.phases[0]'].isna().sum())/volume
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        with open(fileout, "w") as f:
            f.write(str(throat_number))

class EulerDensityII(BasicPNMMetric):
    """
    Class describing Euler density II metric.
    """     
    def __init__(self, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False.             
        """
        super().__init__(vectorizer=None, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 's'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates Euler density for a specific subsample.
        
        **Input:**
        
            cut (numpy.ndarray): 3D array representing a subsample;
        
            cut_name (str): name of subsample;
            
            outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.  
        """
        dimx = cut.shape[0]
        dimy = cut.shape[1]
        dimz = cut.shape[2]
        volume = dimx*dimy*dimz
        df = super().generate(cut_name, gendatadir)
        euler_number = (df['throat.phases[0]'].isna().sum()-df['pore.phase'].isna().sum())/volume
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        with open(fileout, "w") as f:
            f.write(str(euler_number))

class MeanPoreRadius(BasicPNMMetric):
    """
    Class describing mean pore radius metric.
    """ 
    def __init__(self, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False.             
        """
        super().__init__(vectorizer=None, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 's'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates pore radius for a specific subsample.
        
        **Input:**
        
            cut (numpy.ndarray): 3D array representing a subsample;
        
            cut_name (str): name of subsample;
            
            outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.   
        """
        df = super().generate(cut_name, gendatadir)
        mean_pore_radius = np.mean(df['pore.inscribed_diameter'].dropna().tolist())/2
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        with open(fileout, "w") as f:
            f.write(str(mean_pore_radius))

class MeanThroatRadius(BasicPNMMetric):
    """
    Class describing mean throat radius metric.
    """     
    def __init__(self, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False.             
        """
        super().__init__(vectorizer=None, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 's'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates throat radius for a specific subsample.
        
        **Input:**
        
            cut (numpy.ndarray): 3D array representing a subsample;
        
            cut_name (str): name of subsample;
            
            outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.    
        """
        df = super().generate(cut_name, gendatadir)
        mean_throat_radius = np.mean(df['throat.inscribed_diameter'].dropna().tolist())/2
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        with open(fileout, "w") as f:
            f.write(str(mean_throat_radius))
            
class MeanConnectivity(BasicPNMMetric):
    """
    Class describing mean connectivity metric.
    """     
    def __init__(self, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False.             
        """
        super().__init__(vectorizer=None, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 's'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates mean connectivity for a specific subsample.
        
        **Input:**
        
            cut (numpy.ndarray): 3D array representing a subsample;
        
            cut_name (str): name of subsample;
            
            outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.    
        """
        df = super().generate(cut_name, gendatadir)
        pore_number = df.shape[0] - df['pore.phase'].isna().sum()
        con1 = df['throat.conns[0]'].dropna().tolist()
        con2 = df['throat.conns[1]'].dropna().tolist()
        con = con1 +con2
        (unique, counts) = np.unique(con, return_counts=True)
        counts0 = [0 for i in range(pore_number) if i not in unique]
        counts = counts.tolist() + counts0
        mean_con = np.mean(counts)

        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        with open(fileout, "w") as f:
            f.write(str(mean_con))

class PoreRadius(BasicPNMMetric):
    """
    Class describing pore radius metric.
    """ 
    def __init__(self, vectorizer, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            vectorizer (HistVectorizer object): vectorizer to be used for a vector metric;
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False. 
        """
        super().__init__(vectorizer, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates pore radius distribution for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.  
        """
        df = super().generate(cut_name, gendatadir)
        pore_radius = np.array(df['pore.inscribed_diameter'].dropna().tolist())/2
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        np.savetxt(fileout, pore_radius, delimiter='\t')

    def show(self, inputdir, step, cut_id, nbins):
        """
        Vizualize pore radius distribution for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	 
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index;
        	
        	nbins (int): number of bins in histogram.  
        """        
        super().show(inputdir, step, cut_id, nbins, 'pore radius')
        
class ThroatRadius(BasicPNMMetric):
    """
    Class describing throat radius metric.
    """ 
    def __init__(self, vectorizer, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            vectorizer (HistVectorizer object): vectorizer to be used for a vector metric;
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False. 
        """
        super().__init__(vectorizer, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates throat radius distribution for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.    
        """
        df = super().generate(cut_name, gendatadir)
        throat_radius = np.array(df['throat.inscribed_diameter'].dropna().tolist())/2
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        np.savetxt(fileout, throat_radius, delimiter='\t')

    def show(self, inputdir, step, cut_id, nbins):
        """
        Vizualize throat radius distribution for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	 
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index;
        	
        	nbins (int): number of bins in histogram.
        """       
        super().show(inputdir, step, cut_id, nbins, 'throat radius')
        
class Connectivity(BasicPNMMetric):
    """
    Class describing connectivity metric.
    """     
    def __init__(self, vectorizer, n_threads = 1, resolution = 1., show_time = False):
        """
        **Input:**
        
            vectorizer (HistVectorizer object): vectorizer to be used for a vector metric;
        
            n_threads (int): number of threads used for data generation, default: 1;
        
            resolution (float): resolution of studied sample, default: 1;
            
            show_time (bool): Added to monitor time cost for large images,  default: False. 
        """
        super().__init__(vectorizer, n_threads = n_threads, resolution = resolution, show_time = show_time)
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir):
        """
        Generates connectivity distribution for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder;
        	
        	gendatadir (str): folder with generated PNM data.    
        """
        df = super().generate(cut_name, gendatadir)
        pore_number = df.shape[0] - df['pore.phase'].isna().sum()
        con1 = df['throat.conns[0]'].dropna().tolist()
        con2 = df['throat.conns[1]'].dropna().tolist()
        con = con1 +con2
        (unique, counts) = np.unique(con, return_counts=True)
        counts0 = [0 for i in range(pore_number) if i not in unique]
        connectivity = counts.tolist() + counts0
        cut_name_out = cut_name + ".txt"
        fileout = os.path.join(outputdir, cut_name_out)
        np.savetxt(fileout, connectivity, delimiter='\t')

    def show(self, inputdir, step, cut_id, nbins):
        """
        Vizualize the connectivity distribution for a specific subsample.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
        	 
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index;
        	
        	nbins (int): number of bins in histogram. 
        """        
        super().show(inputdir, step, cut_id, nbins, 'connectivity')


# -*- coding: utf-8 -*-
"""
Formulas for the calculation of dREV and sREV sizes for scalar and vector metrics.
"""

def _delta(a, b):
    if (a+b != 0):
        return 2*abs((a-b)/((a+b)))
    else:
        return 0


def get_sREV_size(sigmas_n, threshold):
    """
    sREV size calculation (can be applied both for scalar and vector metrics).
    
    **Input:**
         
    	sigmas_n (dict, int[float]): dictionary, in which a key is a subsamples selection step, and a value is a normailzed standard deviation for this step;
    
    	threshold (float, <1): threshold to estimate sREV size.
    
    **Output:**
    
    	sREV step (int): subsamples selection step, corresponding to sREV size.
    """
    sigmas_n = {key: val for key, val in sigmas_n.items() if val != 0.0}
    steps = list(sigmas_n.keys())
    steps.sort(reverse=True)
    for i, l in enumerate(steps):
        if sigmas_n[l] > threshold:
            if i == 0:
                return None
            else:
                return steps[i-1]
    return steps[-1]


def get_dREV_size_1_scalar(values, threshold):
    """
    dREV size calculation for scalar metric using formula (dREV_size_1). See the documentation.
    
    **Input:**
    
    	values (dict, int[float]): dictionary, in which a key is a subsamples selection step, and a value is a difference of metric mean values for a current and next steps;
      
    	threshold (float, <1): threshold to estimate dREV size.
    
    **Output:**
    
    	dREV step (int): subsamples selection step, corresponding to dREV size.
    """
    steps = list(values.keys())
    steps.sort(reverse=True)
    for i in range(len(steps)-1):
        if _delta(values[steps[i]], values[steps[i+1]]) > threshold:
            if i == 0:
                return None
            else:
                return steps[i]
    return steps[-1]


def get_dREV_size_1_scalar_dimensional(values, threshold):
    """
    dREV size calculation for scalar metric defined i 'x', 'y' an 'z' directions using formula (dREV_size_1).
    Returns subsamples selection step, corresponding to maximal dREV size over all the directions.
    
    **Input:**
    
    	values (dict, int[list(dtype=float)]): dictionary, in which a key is a a subsamples selection step, and a value is a list of the differences of metric mean values for a current and next steps computed in all directions;
    	  
    	threshold (float, <1): threshold to estimate dREV size.
    
    **Output:**
    
    	dREV step(int): subsamples selection step, corresponding to dREV size.
    """
    steps = list(values.keys())
    steps.sort(reverse=True)
    result = []
    for k in range(3):
        label = 0
        for i in range(len(steps)-1):
            if _delta(values[steps[i]][k], values[steps[i+1]][k]) > threshold:
                if i == 0:
                    return None
                else:
                    result.append(steps[i])
                    label = 1
        if (label == 0):
            result.append(sizes[-1])
    return max(result)


def get_dREV_size_2_scalar(values, threshold):
    """
    dREV size calculation for scalar metric using formula (dREV_size_2). See the documentation.
    
    **Input:**
    
    	values (dict, int[float]): dictionary, in which a key is a subsamples selection step, and a value is a difference of metric mean values for a current and next steps;
     
    	threshold (float, <1): threshold to estimate dREV size.
    
    **Output:**
    
    	dREV step (int): subsamples selection step, corresponding to dREV size.
    """
    steps = list(values.keys())
    steps.sort(reverse=True)
    value0 = values[steps[0]]
    for i in range(1, len(steps)):
        if _delta(values[steps[i]], value0) > threshold:
            if i == 1:
                return None
            else:
                return steps[i-1]
    return steps[-1]


def get_dREV_size_2_scalar_dimensional(values, threshold):
    """
    dREV size calculation for scalar metric defined i 'x', 'y' an 'z' directions using formula (dREV_size_2).
    Returns subsamples selection step, corresponding to maximal dREV size over all the directions.
    
    **Input:**
    
    	values (dict, int[list(dtype=float)]): dictionary, in which a key is a subsamples selection step, and a value is a list of the differences of metric mean values for a current and next steps computed in all directions;
    	  
    	threshold (float, <1): threshold to estimate dREV size.
    
    **Output:**
    
    	dREV step (int): subsamples selection step, corresponding to dREV size.
    """
    steps = list(values.keys())
    steps.sort(reverse=True)
    result = []
    for k in range(3):
        value0 = values[steps[0]][k]
        label = 0
        for i in range(len(steps)-1):
            if _delta(values[steps[i]][k], value0) > threshold:
                if i == 0:
                    return None
                else:
                    result.append(steps[i])
                    label = 1
        if label == 0:
            result.append(steps[-1])
    return max(result)


def get_dREV_size_1_vector(values, threshold):
    """
    dREV size calculation for vector metric. 
    
    **Input:**
    
    	values (dict, int[float]): dictionary, in which a key is a subsamples selection step, and a value is a mean distance between vectors at a current and next steps;
    	  
    	threshold (float, <1): threshold to estimate dREV size.
    
    **Output:**
    
    	dREV size (int): subsamples selection step, corresponding to dREV size.
    """
    steps = list(values.keys())
    steps.sort(reverse=True)
    for i in range(len(steps)):
        if values[steps[i]] > threshold:
            if i == 0:
                return None
            else:
                return steps[i-1]
    return steps[-1]



# -*- coding: utf-8 -*-
""" Classes describing metrics to be analyzed."""

from .basic_metric import BasicMetric
from .porosity import Porosity
from .permeability import Permeability
from .euler_density_i import EulerDensityI
from .pnm import BasicPNMMetric, PoreNumber, ThroatNumber, EulerDensityII, MeanPoreRadius, MeanThroatRadius, MeanConnectivity, PoreRadius, ThroatRadius, Connectivity
from .cf import L2, S2, C2, SS, SV, ChordLength, PoreSize
from .pd import BasicPDMetric, PD0, PD1, PD2




# -*- coding: utf-8 -*-
"""Definition of PD-based metrics. For the definition of persistence diagrams (PD) see the documentation."""

from .basic_metric import BasicMetric
import numpy as np
import matplotlib.pyplot as plt
import time
import os
import pyperspairdiamorse as pppdm
from ..vectorizers import SimpleBinningVectorizer, PersistenceImageVectorizer, LandscapeVectorizer, SilhouetteVectorizer


class BasicPDMetric(BasicMetric):
    """
    Base class of PD-based metrics. (Don't use it directly but derive from it).
    """ 
    def __init__(self, vectorizer, n_threads, show_time):
        """
        **Input:**
        
        	vectorizer (SimpleBinningVectorizer, PersistenceImageVectorizer, LandscapeVectorizer or SilhouetteVectorizer object): vectorizer to be used for PD metric.
            
            n_threads (int): number of threads used for data generation;
        	
        	show_time (bool): flag to monitor time cost for large images.
        """
        if not (isinstance(vectorizer, SimpleBinningVectorizer) or isinstance(vectorizer, PersistenceImageVectorizer) or isinstance(
            vectorizer, LandscapeVectorizer) or isinstance(vectorizer, SilhouetteVectorizer)):
            raise TypeError('Vectorizer should be an object of SimpleBinningVectorizer, PersistenceImageVectorizer, LandscapeVectorizer or SilhouetteVectorizer class')
        super().__init__(vectorizer, n_threads = n_threads)
        self.show_time = show_time
        
    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates PD metric for a specific subsample.
        
        **Input:**
        
        	cut (numpy.ndarray): 3D array representing a subsample;
        	
        	cut_name (str): name of subsample;
        	
        	outputdir (str): output folder;
        """ 
        start_time = time.time()
        cut = cut.astype(bool)
        pds = pppdm.extract(cut)
        cut_name_out = cut_name + ".txt"
        for i, elem in enumerate(outputdir):
            fileout = os.path.join(elem, cut_name_out)
            np.savetxt(fileout, pds[i])
        if self.show_time:
            print("cut ", cut_name, ", run time: ")
            print("--- %s seconds ---" % (time.time() - start_time))        

    def show(self, inputdir, step, cut_id, title):
        """
        Transforms generated PD data to the convenient fomat for the following visualization in subclasses.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subsamples;
                	
        	step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index;
            
            title (str): image title.
        """          
        data = self.read(inputdir, step, cut_id)
        _show_pd(data, title)

    def vectorize(self, v1, v2):
        """
        Vectorize the vector metric values for a given pair of subsamples using the method of vectorizer. 
        
        **Input:**
        
        	v1 (list(dtype = float)): data for the first subsample;
        	
        	v2 (list(dtype = float)): data for the second subsample.
        
        **Output:**
        
        	(list(dtype = float), list(dtype = float), float) - a tuple, in which the first two elements are vectorized metric values for a given pair of subsamples, and the last one is the normalized distance between these vectors. 
        """
        return self.vectorizer.vectorize(v1, v2)


class PD0(BasicPDMetric):
    """
    Class describing metric PD of rank 0.
    """    
    def __init__(self, vectorizer, n_threads = 1, show_time = False):
        """
        **Input:**
        
        	vectorizer (SimpleBinningVectorizer, PersistenceImageVectorizer, LandscapeVectorizer or SilhouetteVectorizer object): vectorizer to be used for PD metric;
            
            n_threads (int): number of threads used for data generation;
        	
        	show_time (bool): flag to monitor time cost for large images.        
        """
        super().__init__(vectorizer, n_threads, show_time)
        self.metric_type = 'v'
        
    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the PD of rank 0 for a specific subcube.
        
        **Input:**
        
        	cut (numpy.ndarray): subsample;
        	
        	cut_name (str): name of subcube;
        	
        	outputdir (str): output folder.
        """
        super().generate(cut, cut_name, outputdir)

    def show(self, inputdir, cut_step, cut_id):
        """
        Vizualize the PD of rank 0 for a specific subcube.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subcubes;
        	 
        	cut_step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index.
        """  
        title = 'PD0' + ",  step = " + str(cut_step) + ", id = " + str(cut_id)
        super().show(inputdir, cut_step, cut_id, title)

class PD1(BasicPDMetric):
    """
    Class describing metric PD of rank 1.
    """ 
    def __init__(self, vectorizer, n_threads = 1, show_time = False):
        """
        **Input:**
        
        	vectorizer (SimpleBinningVectorizer, PersistenceImageVectorizer, LandscapeVectorizer or SilhouetteVectorizer object): vectorizer to be used for PD metric;
            
            n_threads (int): number of threads used for data generation;
        	
        	show_time (bool): flag to monitor time cost for large images.        
        """
        super().__init__(vectorizer, n_threads, show_time)
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the PD of rank 1 for a specific subcube.
        
        **Input:**
        
        	cut (numpy.ndarray): subcube;
        	
        	cut_name (str): name of subcube;
        	
        	outputdir (str): output folder.
        """
        super().generate(cut, cut_name, outputdir)

    def show(self, inputdir, cut_step, cut_id):
        """
        Vizualize the PD of rank 0 for a specific subcube.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subcubes;
        	 
        	cut_step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index.
        """  
        title = 'PD1' + ",  step = " + str(cut_step) + ", id = " + str(cut_id)
        super().show(inputdir, cut_step, cut_id, title)


class PD2(BasicPDMetric):
    """
    Class describing metric PD of rank 2.
    """ 
    def __init__(self, vectorizer, n_threads = 1, show_time = False):
        """
        **Input:**
        
        	vectorizer (SimpleBinningVectorizer, PersistenceImageVectorizer, LandscapeVectorizer or SilhouetteVectorizer object): vectorizer to be used for PD metric;
            
            n_threads (int): number of threads used for data generation;
        	
        	show_time (bool): flag to monitor time cost for large images.        
        """
        super().__init__(vectorizer, n_threads, show_time)
        self.metric_type = 'v'

    def generate(self, cut, cut_name, outputdir, gendatadir = None):
        """
        Generates the PD of rank 2 for a specific subcube.
        
        **Input:**
        
        	cut (numpy.ndarray): subcube;
        	
        	cut_name (str): name of subcube;
        	
        	outputdir (str): output folder.
        """
        super().generate(cut, cut_name, outputdir)

    def show(self, inputdir, cut_step, cut_id):
        """
        Vizualize the PD of rank 0 for a specific subcube.
        
        **Input:**
        
        	inputdir (str): path to the folder containing generated metric data for subcubes;
        	 
        	cut_step (int): subsamples selection step;
        	
        	cut_id (int: 0,..8): cut index.
        """  
        title = 'PD2' + ",  step = " + str(cut_step) + ", id = " + str(cut_id)
        super().show(inputdir, cut_step, cut_id, title)


def _show_pd(data, title):
    b = [elem[0] for elem in data]
    d = [elem[1] for elem in data]
    plt.rcParams.update({'font.size': 16})
    plt.rcParams['figure.dpi'] = 300
    fig, ax = plt.subplots()
    plt.title(title)
    plt.plot(b, d, "ro")
    plt.axhline(y=0, color='black', linestyle='-')
    plt.axvline(x=0, color='black', linestyle='-')
    plt.xlabel("birth")
    plt.ylabel("death")
    plt.show()
